

boolean[][] cellsNow; //2 dimensional array of bools
boolean[][] cellsFuture; //2 dimensional array of bools


float cellW=5; //the size of the trangles
float cellH= (sqrt(3.0)/2.0)*cellW; //from solving Pithagoras...
int columns, rows;

//these are the rules for a triangular
//(there are a maximum of 3 neighbours on).
//There going to be then a maximum number of rules 
//of a binary number of with 6 bits, or 2^6, which is
//63...
int currentRule=0;

boolean[] trueRules=new boolean[4];
boolean[] falseRules=new boolean[4];

void setup(){

size(1000,500);
pixelDensity(displayDensity());
background(255); //draw background white

frameRate(4);

noStroke();

//depending of size
rows=int(height/cellH); 
//columns=int(2*width/cellW)-1;
columns=int(2*(2*width-rows*cellW)/(2*cellW))-1;

if(columns<=0){
    println("make screen wider or smaller in height. Cannot fit array");
    columns=0;

}else if(columns%2!=0){ //if not even, it does not work either, because the wrapping does not make sense
    columns--;

}

cellsNow=new boolean[columns][rows]; 
cellsFuture=new boolean[columns][rows];

makeRulesFromNumber(currentRule);
initialise();

}

void draw(){

  background(255); //clean the background
  
  drawCA(); //draw
  
   tick(); //calculate future values

   update(); 
}

// we draw them... so each 
void drawCA(){
    //noStroke();
    for(int i=0;i<columns;i++){//we go through all the columns
      for(int j=0;j<rows;j++){

        //the colour:
        if(cellsNow[i][j]){
            fill(0);
        }else{
            fill(255);
        }
        beginShape();

        //Draw Triangular
        //horizontal shift...it is hard to see how this works without
        //drawing it...
        float hShift=((rows-j)*cellW)/2.0; 
        int rectPos=i/2;

        if(i%2==0) //if even, triangle with base down
        {
            vertex(rectPos*cellW+hShift, (j+1)*cellH);
            vertex((rectPos+1)*cellW+hShift, (j+1)*cellH);
            vertex((rectPos+0.5)*cellW+hShift, j*cellH);
        }else{ //triangle with base up
            vertex((rectPos+1  )*cellW+hShift, (j+1)*cellH);
            vertex((rectPos+1.5)*cellW+hShift, j*cellH);
            vertex((rectPos+0.5)*cellW+hShift, j*cellH);
        } 
        
        
        //Draw rectangular (simpler):
        // int rectPos=i/2;
        // if(i%2==0) //if even, triangle with base down
        // {
        //     vertex(rectPos*cellW, (j+1)*cellH);
        //     vertex((rectPos+1)*cellW, (j+1)*cellH);
        //     vertex(rectPos*cellW, j*cellH);
        // }else{ //triangle with base up
        //     vertex((rectPos+1)*cellW, (j+1)*cellH);
        //     vertex((rectPos+1)*cellW, j*cellH);
        //     vertex(rectPos*cellW, j*cellH);
        // } 
         
        

        endShape(CLOSE);



      }
    }

}


void tick(){
  for(int i=0;i<columns;i++){//we go through all the columns
    for(int j=0;j<rows;j++){
      int numNeigs= countNeighbours(i,j);
       cellsFuture[i][j]=applyRules(cellsNow[i][j],numNeigs);
    }
  }
}

//this is the neighborhood of a triangular CA:
//
//                    __  __  __  __
//                  /\  /\  /\  /\  /
//                 /__\/__\/__\/__\/
//                /\  /\  /\  /\  /
//               /__\/__\/__\/__\/
//              /\  /\  /\  /\  /
//             /__\/__\/__\/__\/
//            /\  /\  /\  /\  /
//           /__\/__\/__\/__\/  
//    
//   Each triangle in each row will have different neighbours,
//   depending if its base is up or down. This will correspond
//   with the fact of then having even or odd indices in the row
//   
//   So the neighborhoods will be:
//   
//   For EVEN indices               And for ODD indices
//   (since they start with 0) 
//   
//                                      / \
//                                     / * \
//       ______ ______                /_ |_ \ 
//       \    / \    /               /\  |  /\ 
//        \ *--*--* /               / *--*--* \  
//         \/_ | _\/               /____\ /____\        
//          \  |  /                  
//           \ * /                
//            \ /
//
//
//    Since the triangles are really in a rectangular array, it becomes this:

//
//   --- -- -- -- -- 
//   |\1|\3|\5|\7|\9|
//   |0\|2\|4\|6\|8\|
//   --- -- -- -- -- 
//   |\1|\3|\5|\7|\9|
//   |0\|2\|4\|6\|8\|
//   --- -- -- -- -- 
// 
//     
//   So the indices will be:
//                   *
//                   |
//   *--#--*   and   *--#--*
//         |
//         *
//         
//   for example cell 2 in the top row will have as its neighbours in columns  1 and 3 
//   in the same row, and in column 3 in the row below. Cell 3 in the second row will have
//   columns 2 and 4 in the same row as its neighbours and cell in column 2 in the  top row 

int countNeighbours(int col,int row){

    // % is the remainder operator. 
    // If the reminder of col divided by 2 is 0
    // It means that the column is even (it can be divided by 2)
     
    int  leftNeig=wrap(col-1,columns);
    int rightNeig=wrap(col+1,columns);

    int colNeig, rowNeig;

    if(col % 2 == 0){ 
        rowNeig=wrap(row+1,rows);
        colNeig=rightNeig;
    }else{ //odd
        rowNeig=wrap(row-1,rows);
        colNeig=leftNeig;
    }


    int totalNeighbours=0;

    //same row:
    if(cellsNow[leftNeig][row]==true){
        totalNeighbours++;
    }

    if(cellsNow[rightNeig][row]==true){
        totalNeighbours++;
    }

    //the other one, on top or below, depending of the case
    if(cellsNow[colNeig][rowNeig]==true){
            totalNeighbours++;
    }

  return totalNeighbours;
}

//this is the wrap function.
int wrap(int index,int maxBound){
  int result;
  if(index<0){
     result=maxBound+index;
  }
  else if(index>=maxBound){
    result=index-maxBound;
  }
  else{
    result=index;
  }
  return result;
}


boolean applyRules(boolean me, int numNeigs){
  if(me==true){
    return trueRules[numNeigs];
  }else {
    return falseRules[numNeigs];
  } 
}


void update(){
  //copying all values.
  for(int i=0;i<columns;i++){//we go through all the columns
    for(int j=0;j<rows;j++){
      cellsNow[i][j]=cellsFuture[i][j];
    }
  }
}


void makeRulesFromNumber(int rule){ //generate rules from Wolfram-equivalent encoding
  for(int i=0;i<4;i++){
    falseRules[i] = (rule & 1<<(2*i))!=0;
    trueRules[i]  = (rule & 1<<(2*i+1))!=0;
  }
  println("//Rule "+rule);
  println("boolean[] trueRules={");
  for(int i=0;i<4;i++){
    println("\t"+trueRules[i] +",");
  }
  println("};");
  println("boolean[] falseRules={");
  for(int i=0;i<4;i++){
    println("\t"+falseRules[i] +",");
  }
  println("};");
}

void initialise(){

    for(int i=0;i<columns;i++){//we go through all the columns
      for(int j=0;j<rows;j++){
        cellsNow[i][j]=false;
      }
    }

    int centreCol=columns/2+1;
    int centreRow=rows/2;

    cellsNow[centreCol][centreRow]=true;
    cellsNow[centreCol+1][centreRow]=true;
    cellsNow[centreCol-1][centreRow]=true;

    if((centreCol) % 2 == 0){ 
        cellsNow[centreCol+1][centreRow+1]=true;
    }else{ //odd
        cellsNow[centreCol-1][centreRow-1]=true;
    }
}


void keyPressed(){
    if(key=='2'){
        currentRule++;
        currentRule=currentRule>=64 ? 0 : currentRule;
        makeRulesFromNumber(currentRule);
        initialise();

    }else if(key=='1'){
        currentRule--;
        currentRule=currentRule<0 ? 63 : currentRule;
        makeRulesFromNumber(currentRule);
        initialise();
    }

}